import asyncio
import logging
from typing import AsyncIterator

import aiohttp
from apscheduler.schedulers.async_ import AsyncScheduler
from apscheduler.triggers.interval import IntervalTrigger

from fastchecks import conf, require, util, vutil
from fastchecks.check import check_website
from fastchecks.sockets import CheckResultSocket, WebsiteCheckSocket
from fastchecks.sockets.postgres import CheckResultSocketPostgres, WebsiteCheckSocketPostgres
from fastchecks.types import CheckResult, WebsiteCheck, WebsiteCheckScheduled

# -----------------------------------------------------------------------------


class ChecksRunnerContext:
    """
    Holds the state context to be able to run checks and store results.

    The backends/sockets for checks and results could potentially be different (e.g. Redis or even a simple text file for checks, Postgres for results).
    """

    def __init__(
        self,
        session: aiohttp.ClientSession,
        checks: WebsiteCheckSocket,
        results: CheckResultSocket,
        default_interval_seconds: int | None = None,
    ) -> None:
        require(not session.closed, "Session must be open")
        require(not checks.is_closed(), "Checks socket must be open")
        require(not results.is_closed(), "Results socket must be open")

        self._aiohttp_session = session
        self.checks = checks
        self.results = results
        self.default_interval_seconds = (
            conf.DEFAULT_CHECK_INTERVAL_SECONDS
            if default_interval_seconds is None
            else conf.validated_interval(default_interval_seconds)
        )
        """Default interval for website checks that don't specify it"""

    # -----------------------------------------------------------------------------

    @classmethod
    def init_with_postgres(cls, postgres_conninfo: str, **kwargs) -> "ChecksRunnerContext":
        vutil.validated_postgres_conninfo(postgres_conninfo)

        return cls(
            session=aiohttp.ClientSession(),
            checks=WebsiteCheckSocketPostgres(postgres_conninfo),
            results=CheckResultSocketPostgres(postgres_conninfo),
            **kwargs,
        )

    # -----------------------------------------------------------------------------

    async def __aenter__(self) -> "ChecksRunnerContext":
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        await asyncio.gather(self._aiohttp_session.close(), self.checks.close(), self.results.close())

    async def close(self) -> None:
        await self.__aexit__(None, None, None)

    # -----------------------------------------------------------------------------

    def get_interval_seconds(self, check: WebsiteCheckScheduled) -> int:
        return self.default_interval_seconds if check.interval_seconds is None else check.interval_seconds

    # -----------------------------------------------------------------------------

    async def check_only(self, check: WebsiteCheck) -> CheckResult:
        """Check website without saving into results data storage."""
        ret = await check_website(self._aiohttp_session, check)
        logging.debug(ret)
        return ret

    async def check_n_write(self, check: WebsiteCheck) -> CheckResult:
        """Check website and save into results data storage."""
        ret = await self.check_only(check)
        await self.results.write(ret)
        return ret

    async def check_all_once_n_write(self) -> AsyncIterator[CheckResult]:
        async for check in self.checks.read_n(util.PRACTICAL_MAX_INT):
            result = await self.check_only(check)
            await self.results.write(result)
            yield result

    # -----------------------------------------------------------------------------

    async def _add_check_to_scheduler(self, scheduler: AsyncScheduler, check: WebsiteCheckScheduled) -> AsyncScheduler:
        fun = self.check_n_write

        # Note: we can later retrieve scheduled checks by their url (with `AsyncScheduler.get_schedule``)
        # MAYBE (2023-07-09; future idea): tag the check with the url's domain, so later we can filter on them
        await scheduler.add_schedule(
            fun, trigger=IntervalTrigger(seconds=self.get_interval_seconds(check)), id=check.url, args=[check]
        )

        return scheduler

    async def _add_checks_to_scheduler(
        self, scheduler: AsyncScheduler, checks: AsyncIterator[WebsiteCheckScheduled]
    ) -> AsyncScheduler:
        async for check in checks:
            await self._add_check_to_scheduler(scheduler, check)

        return scheduler

    async def run_checks_until_stopped_in_foreground(self) -> None:
        """
        Run all saved checks until stopped (e.g. with Ctrl+C) in the foreground.
        The interval for each check is taken from the check itself, or the default interval if not specified.

        If there are no checks yet, an error is raised.
        """
        try:
            async with AsyncScheduler() as scheduler:
                default_interval_msg = f" -- with default interval: {self.default_interval_seconds}s"

                async for check in await self.checks.read_all():
                    await self._add_check_to_scheduler(scheduler, check)
                    print(f"Adding check to scheduler: {check}{'' if check.interval_seconds else default_interval_msg}")

                require(len(await scheduler.get_schedules()) != 0, "No checks to run. Add some checks first.")

                print("\nRunning until stopped...")
                await scheduler.run_until_stopped()

        except (KeyboardInterrupt, SystemExit):
            pass



import datetime
from typing import AsyncIterator, TypeVar
from urllib.parse import urlparse, urlunparse
import sys
import ctypes

import aiohttp


PRACTICAL_MAX_INT = sys.maxsize // ctypes.sizeof(ctypes.c_void_p)
"""
A practical maximum value for an integer in Python.
It's handy, for instance, to use it when you want to read all the results from a socket.

Value:
* 32bit machine: (2**31 - 1)//4 = 1152921504606846975
* 64bit machine: (2**63 - 1)//8 = 1152921504606846975

See:
https://docs.python.org/2/library/sys.html#sys.maxsize
https://stackoverflow.com/questions/855191/how-big-can-a-python-list-get#comment112727918_15739630
https://stackoverflow.com/a/1406210/341320
"""


def str_pad(c: int, size: int = 3) -> str:
    return str(c).zfill(size)


def shorten_str(x: str, max=100) -> str:
    if len(x) > max:
        x = x[:max] + "..."
    return x


def replace_url_last_segment(url: str, new_segment: str) -> str:
    """
    Example behavior (new_segment = "fastchecks"), urls:
    * "postgresql://localhost/postgres" -> "postgresql://localhost/fastchecks"
    * "postgresql://localhost/postgres?sslmode=require" -> "postgresql://localhost/fastchecks?sslmode=require"
    * "postgresql://username:password@localhost:5432/postgres?sslmode=require" -> "postgresql://username:password@localhost:5432/fastchecks?sslmode=require"
    """
    parsed_url = urlparse(url)
    path_parts = parsed_url.path.split("/")
    path_parts[-1] = new_segment
    modified_path = "/".join(path_parts)

    modified_url = urlunparse(
        (parsed_url.scheme, parsed_url.netloc, modified_path, parsed_url.params, parsed_url.query, parsed_url.fragment)
    )

    return modified_url


def get_utcnow() -> datetime.datetime:
    """
    Return the current UTC timestamp in seconds.

    Use this method to make sure you are always using a timestamp with same timezone (UTC).
    """
    return datetime.datetime.utcnow()


def get_utcnow_time_difference_seconds(timestamp_start: datetime.datetime) -> float:
    """
    Return the time difference in seconds between the current UTC datetime and the input datetime.
    """
    return (get_utcnow() - timestamp_start).total_seconds()


def is_likely_text_based_body(response: aiohttp.ClientResponse) -> bool:
    """
    Return True if the response's content type is likely to be text-based (e.g. HTML, JSON, XML, etc.).

    Note: this method is not 100% reliable, but good enough for our purposes.
    """
    content_type = response.content_type.lower()
    return (
        response.charset is not None
        or content_type.startswith("text/")
        or content_type.endswith("+xml")
        or content_type.endswith("+json")
        or content_type.endswith("+xhtml")
        or content_type.endswith("+html")
    )


def is_content_length_less_than(response: aiohttp.ClientResponse, length: int, allow_none_content_length: bool) -> bool:
    content_length = response.headers.get("Content-Length")

    if content_length is None:
        return allow_none_content_length
    else:
        return int(content_length) < length


_A = TypeVar("_A")


# MAYBE #1 (2023-07-08) improve with real async mapping
async def async_itr_to_list(x: AsyncIterator[_A]) -> list[_A]:
    return [result async for result in x]



#
# Utils for validating data values, specifically user inputs.
# - Further validation functions are defined conf module (which depend on configuration values).
#
# Conventions:
# * Unless otherwise specified, validators must raise ValueError if the value is invalid.
# * Functions that validate a value and (if valid) return it, are prefixed with "validated_".
# * Functions that validate a value and (if valid) return an optionally computed value, are prefixed with "validate_".
#

from numbers import Number
from urllib.parse import ParseResult, urlparse
import re2

from fastchecks import meta, require


def validated_parsed_is_positive_int(val: str) -> int:
    return validated_is_positive_int(int(val))


def validated_is_positive_int(val: int) -> int:
    require(val > 0, f"Value must be positive: {val}")
    return val


def validated_in_range(name: str, val: Number, min: Number, max: Number) -> Number:
    """
    Validate that the given value is between the given min and max, and return it if it is, otherwise raise ValueError.
    """
    require(
        min <= val <= max,
        f"{name} must be between min & max: [{min}, {max}]",
    )
    return val


ACCEPTED_WEB_URL_SCHEMES = {"http", "https"}


def validate_url(url: str, accepted_schemes: set[str], raise_error: bool = True) -> ParseResult | None:
    """
    Validate that the given string is a valid URL and has one of the accepted schemes.
    * If the URL is valid, return the urlparse's parsed result.
    * Else:
        * if raise_error is True, raise ValueError.
        * else, return None.
    """
    # See: https://snyk.io/blog/secure-python-url-validation/

    try:
        ret = urlparse(url)
        accept: bool = bool(ret.scheme) and bool(ret.netloc) and ret.scheme in accepted_schemes
    except:
        return None

    if accept:
        return ret
    elif raise_error:
        raise ValueError(f"Invalid URL (also it must start with a scheme in: {accepted_schemes}): {url}")
    else:
        return None


def validated_web_url(url: str) -> str:
    validate_url(url, accepted_schemes=ACCEPTED_WEB_URL_SCHEMES, raise_error=True)
    return url


ACCEPTED_PG_CONNINFO_URL_SCHEMES = {"postgres", "postgresql"}


def validated_postgres_conninfo(conninfo: str) -> str:
    require(
        validate_url(conninfo, accepted_schemes=ACCEPTED_PG_CONNINFO_URL_SCHEMES, raise_error=False) is not None,
        f"The Postgres conninfo must be of URL form and start with a valid scheme ({ACCEPTED_PG_CONNINFO_URL_SCHEMES}) (e.g. for a local Postgres database, 'postgresql://localhost:5432/{meta.NAME}')",
    )
    return conninfo


def validate_regex(regex: str, raise_error: bool = True) -> re2._Regexp | None:
    """
    Validate regex string: the regex must be compilable with google's re2 library.

    If the regex is valid, return its re2._Regexp for possible later use.
    Else if raise_error is True, raise ValueError.
    """
    try:
        return re2.compile(regex)
    except re2.error:
        if raise_error:
            raise ValueError(
                f"Invalid regex (cannot compile it with google-re2 regex syntax https://github.com/google/re2/wiki/Syntax): {regex}"
            )
        else:
            return None


def validated_regex(regex: str) -> str:
    validate_regex(regex, raise_error=True)
    return regex


def validated_regex_accepting_none(regex: str | None) -> str | None:
    if regex is None:
        return None
    else:
        validate_regex(regex, raise_error=True)
        return regex



import datetime
from pydantic import BaseModel

from fastchecks import util, vutil
from fastchecks import conf, require


# We use Pydantic classes for type safety and because they could be handy in the future for de/serialization.
# See benchmarks with other alternatives (e.g. NamedTuple): https://janhendrikewers.uk/pydantic_vs_protobuf_vs_namedtuple_vs_dataclasses.html
# Note that Pydantic V2 (released in 2023 June) promises a 5-50x speedup over V1: https://docs.pydantic.dev/2.0/blog/pydantic-v2-alpha/#headlines

_MAX_REPR_LEN: int = 500


class WebsiteCheck(BaseModel):
    url: str
    regex: str | None = None

    @classmethod
    def with_validation(cls, url: str, regex: str | None = None) -> "WebsiteCheck":
        """
        Validate the given URL and regex (if any), and return a WebsiteCheck instance.

        If the URL or regex are invalid, raise ValueError.
        """
        return cls(url=vutil.validated_web_url(url), regex=vutil.validated_regex_accepting_none(regex))

    @classmethod
    def without_validation(cls, url: str, regex: str | None = None) -> "WebsiteCheck":
        """
        Return a WebsiteCheck instance without validating the given URL and regex (if any).

        Only use this method if you are sure that the URL and regex are valid (e.g., they were validated before).
        """
        return cls(url=url, regex=regex)

    def __repr__(self) -> str:
        return util.shorten_str(super().__repr__(), max=_MAX_REPR_LEN)

    def __str__(self) -> str:
        return util.shorten_str(super().__repr__(), max=_MAX_REPR_LEN)


class WebsiteCheckScheduled(WebsiteCheck):
    """
    WebsiteCheck with an schedule.
    """

    # url: str
    # regex: str | None = None
    interval_seconds: int | None
    """If None, later a system's default value will be used."""

    @classmethod
    def with_check(cls, check: WebsiteCheck, interval_seconds: int | None) -> "WebsiteCheckScheduled":
        return cls(
            url=check.url, regex=check.regex, interval_seconds=conf.validated_interval_accepting_none(interval_seconds)
        )


class CheckResult(BaseModel):
    #
    check: WebsiteCheck
    """The check that generated this result"""
    #
    # Common values
    #
    timestamp_start: datetime.datetime
    response_time: float
    #
    # Connection error values
    #
    timeout_error: bool
    host_error: bool
    other_error: bool
    #
    # Response values (note: OK response, <400, or not)
    #
    response_status: int | None
    regex_match: str | bool | None
    """
    str: the tested regex matched and the matched string is this value.
    bool True: the tested regex matched but the matched string is not available.
    bool False: the tested regex did not match.
    None: means "not tested" (e.g. there was no regex to test, the response was not OK, or the response body was ignored because it was too big or not text).
    """

    def __init__(self, **data) -> None:
        """
        Validate the input data.
        """
        super().__init__(**data)

        if self.check.regex is None:
            require(self.regex_match is None, "If there is no regex, regex_match MUST be None.")

    def __repr__(self) -> str:
        return util.shorten_str(super().__repr__(), max=_MAX_REPR_LEN)

    def __str__(self) -> str:
        return util.shorten_str(super().__repr__(), max=_MAX_REPR_LEN)

    def is_success(self) -> bool:
        """
        Return True if the check was successful (i.e. no error, response is OK, and expected regex wax matched if any).
        """
        return self.is_response_ok() and self.is_regex_validated()

    def is_response_ok(self) -> bool:
        """
        Return True if the check was partially successful (i.e. no error and response is OK).
        """
        return self.response_status is not None and self.response_status < 400

    def is_regex_validated(self) -> bool:
        """
        Return True if the regex was validated (i.e. there was a regex and it matched).
        """
        return self.check.regex is None or self.is_regex_match_truthy()

    def is_regex_match_truthy(self) -> bool:
        """
        Return True if the regex matched and it's either True or the matched string is present.

        Note: we consider a regex match to be truthy even if the matched string is the empty string.
        This could be changed.
        """
        return self.regex_match is True or isinstance(self.regex_match, str)

    def regex_match_to_bool_or_none(self) -> bool | None:
        return None if self.regex_match is None else self.is_regex_match_truthy()

    @classmethod
    def response(
        cls,
        check: WebsiteCheck,
        timestamp_start: datetime.datetime,
        response_time: float,
        response_status: int,
        regex_match: str | bool | None,
    ) -> "CheckResult":
        """
        Return a successful CheckResult.
        """
        return cls(
            check=check,
            #
            timestamp_start=timestamp_start,
            response_time=response_time,
            #
            timeout_error=False,
            host_error=False,
            other_error=False,
            #
            response_status=response_status,
            regex_match=regex_match,
        )

    @classmethod
    def failure(
        cls,
        check: WebsiteCheck,
        timestamp_start: datetime.datetime,
        response_time: float,
        timeout_error: bool = False,
        host_error: bool = False,
        other_error: bool = False,
    ) -> "CheckResult":
        """
        Return a failed CheckResult.
        """
        require(
            (timeout_error or host_error or other_error) and not (timeout_error and host_error and other_error),
            "There can only be one error type.",
        )

        return cls(
            check=check,
            #
            timestamp_start=timestamp_start,
            response_time=response_time,
            #
            timeout_error=timeout_error,
            host_error=host_error,
            other_error=other_error,
            #
            response_status=None,
            regex_match=None,
        )



from abc import ABC, abstractmethod
from typing import AsyncIterator
from pydantic.types import PositiveInt

from fastchecks.types import WebsiteCheckScheduled, CheckResult
from fastchecks.util import PRACTICAL_MAX_INT


class WebsiteCheckSocket(ABC):
    @abstractmethod
    def is_closed(self) -> bool:
        ...

    @abstractmethod
    async def upsert(self, check: WebsiteCheckScheduled) -> None:
        """
        Upsert a check into the socket's underlying storage.
        That is, if the check's URL already exists, update the regex.
        Otherwise, insert/add/write the check.
        """
        ...

    @abstractmethod
    async def read_n(self, n: PositiveInt) -> AsyncIterator[WebsiteCheckScheduled]:
        ...

    async def read_all(self) -> AsyncIterator[WebsiteCheckScheduled]:
        """
        Read all checks from the socket's underlying storage.

        This is a sugar method that calls `read_n` with a practically infinite large number.
        Careful: Use this method only if the underlying storage can handle reading all checks (e.g., there not many written checks, which it's common).
        """
        return self.read_n(PRACTICAL_MAX_INT)

    @abstractmethod
    async def delete(self, url: str) -> int:
        """Return number of deleted checks, 0 (no actual deletion) or 1 (did delete)"""
        ...

    @abstractmethod
    async def delete_all(self, confirm: bool) -> int:
        """Return number of deleted checks. If the number is unknown, return a negative number like -1"""
        ...

    @abstractmethod
    async def close(self) -> None:
        ...

    async def __aenter__(self) -> "WebsiteCheckSocket":
        return self

    async def __aexit__(self, exc_type, exc_value, traceback) -> None:
        await self.close()


class CheckResultSocket(ABC):
    @abstractmethod
    def is_closed(self) -> bool:
        ...

    @abstractmethod
    async def write(self, result: CheckResult) -> None:
        ...

    @abstractmethod
    async def read_last_n(self, n: PositiveInt) -> AsyncIterator[CheckResult]:
        ...

    @abstractmethod
    async def close(self) -> None:
        ...

    async def __aenter__(self) -> "CheckResultSocket":
        return self

    async def __aexit__(self, exc_type, exc_value, traceback) -> None:
        await self.close()



import os
from typing import Callable, TypeVar
from fastchecks import vutil

_CONVERSION_OUTPUT = TypeVar("_CONVERSION_OUTPUT")


def get_typed_envar(
    envar_name: str, default: _CONVERSION_OUTPUT, conversion: Callable[[str], _CONVERSION_OUTPUT]
) -> _CONVERSION_OUTPUT:
    """
    Return the value of the given environment variable, or the default value if the environment variable is not set.
    """
    value = os.environ.get(envar_name)
    if value is None:
        return default
    else:
        return conversion(value)


def read_envar_value(envar_name: str, envar_value: _CONVERSION_OUTPUT | None) -> _CONVERSION_OUTPUT:
    """
    Read cached value of the given environment variable, or raise ValueError if the environment variable is not set.
    """
    if envar_value is None:
        raise ValueError(f"Environment variable is not set: {envar_name}")
    else:
        return envar_value


# -----------------------------------------------------------------------------


DEFAULT_REQ_TIMEOUT_SECONDS: float = get_typed_envar(
    "FC_DEFAULT_REQ_TIMEOUT_SECONDS", default=10.0, conversion=lambda x: float(x)
)

MIN_INTERVAL_SECONDS: int = get_typed_envar("FC_MIN_INTERVAL_SECONDS", default=5, conversion=lambda x: int(x))

MAX_INTERVAL_SECONDS: int = get_typed_envar("FC_MAX_INTERVAL_SECONDS", default=300, conversion=lambda x: int(x))


def validated_parsed_interval(interval_seconds: str) -> int:
    return validated_interval(int(interval_seconds))


def validated_interval(interval_seconds: int, name: str = "interval") -> int:
    return vutil.validated_in_range(name, interval_seconds, MIN_INTERVAL_SECONDS, MAX_INTERVAL_SECONDS)


def validated_interval_accepting_none(interval_seconds: int | None, name: str = "interval") -> int | None:
    if interval_seconds is None:
        return None
    else:
        return vutil.validated_in_range(name, interval_seconds, MIN_INTERVAL_SECONDS, MAX_INTERVAL_SECONDS)


DEFAULT_CHECK_INTERVAL_SECONDS: int = validated_interval(
    get_typed_envar("FC_DEFAULT_CHECK_INTERVAL_SECONDS", default=180, conversion=lambda x: int(x)),
    name="FC_DEFAULT_CHECK_INTERVAL_SECONDS",
)


# -----------------------------------------------------------------------------

TOO_BIG_CONTENT_LENGTH_KB = get_typed_envar("FC_TOO_BIG_CONTENT_LENGTH_KB", default=100000, conversion=lambda x: int(x))
"""
Limit value (in _kilo_ bytes, not kibi bytes) to consider a response's content length as too big to be read in memory.

For reference, the size of https://python.org is, as of 2023-07-06, 49943 bytes.
"""

# -----------------------------------------------------------------------------

_POSTGRES_CONNINFO_ENVAR_NAME = "FC_POSTGRES_CONNINFO"

_POSTGRES_CONNINFO: str | None = os.environ.get(_POSTGRES_CONNINFO_ENVAR_NAME)


def get_postgres_conninfo() -> str:
    """
    Return the Postgres conninfo envar value, or raise ValueError if the environment variable is not set or invalid.
    """
    ret = read_envar_value(_POSTGRES_CONNINFO_ENVAR_NAME, _POSTGRES_CONNINFO)
    return vutil.validated_postgres_conninfo(ret)



import tomllib

with open("pyproject.toml", "rb") as f:
    _META = tomllib.load(f)

NAME = _META["tool"]["poetry"]["name"]
VERSION = _META["tool"]["poetry"]["version"]
DESCRIPTION = _META["tool"]["poetry"]["description"]
WEBSITE = _META["tool"]["poetry"].get("homepage", _META["tool"]["poetry"]["repository"])



import argparse
from argparse import Namespace as NamedArgs
import asyncio
import sys
from typing import Any

from fastchecks import conf, util, vutil
from fastchecks.runner import ChecksRunnerContext
from fastchecks.types import WebsiteCheck, WebsiteCheckScheduled
from fastchecks import meta

# ---------------------------------------------------------------------------

#
# Main parser
#

PARSER = argparse.ArgumentParser(
    prog=meta.NAME, description=meta.DESCRIPTION, epilog=f"For more help check: {meta.WEBSITE}"
)
PARSER.add_argument(
    "--pg_conninfo",
    type=vutil.validated_postgres_conninfo,
    help=f"(Default: read from envar {conf._POSTGRES_CONNINFO_ENVAR_NAME}) PostgreSQL connection info",
    default=conf._POSTGRES_CONNINFO,
)


# -----------------------------------------------------------------------------

#
# Common command arguments
#


def _url_kwargs(**kwargs) -> dict[str, Any]:
    return {"type": vutil.validated_web_url, "help": "The URL to check", **kwargs}


def _regex_kwargs(**kwargs) -> dict[str, Any]:
    return {
        "type": vutil.validated_regex,
        "help": "(Default: no check) The regex to match against the response body",
        **kwargs,
    }


def _interval_kwargs(**kwargs) -> dict[str, Any]:
    return {
        "type": conf.validated_parsed_interval,
        "help": f"(Default: {conf.DEFAULT_CHECK_INTERVAL_SECONDS}) The interval in _seconds_ for a check when it is scheduled to be run periodically (min: {conf.MIN_INTERVAL_SECONDS}, max: {conf.MAX_INTERVAL_SECONDS})",
        **kwargs,
    }


# -----------------------------------------------------------------------------

#
# Following subparsers (commands)
#

SUBPARSERS = PARSER.add_subparsers(
    title="Commands",
    dest="command",
    description=" ",
    help="Info:",
)

# -----------------------------------------------------------------------------


def _add_upsert_check(subparsers: argparse._SubParsersAction) -> tuple[argparse._SubParsersAction, Any]:
    cmd = subparsers.add_parser(
        "upsert_check",
        help="Write a new check to the data store, or update an existing check (uniquely identified by its URL)",
    )

    cmd.add_argument("url", **_url_kwargs())
    cmd.add_argument("--regex", **_regex_kwargs())
    cmd.add_argument("--interval", **_interval_kwargs())

    async def fun(ctx: ChecksRunnerContext, x: NamedArgs):
        await ctx.checks.upsert(
            # The args are already validated, but just in case
            WebsiteCheckScheduled.with_check(WebsiteCheck.with_validation(x.url, x.regex), interval_seconds=x.interval)
        )

    cmd.set_defaults(fun=fun)

    return (subparsers, cmd)


_add_upsert_check(SUBPARSERS)


# -----------------------------------------------------------------------------


def _add_read_all_checks(subparsers: argparse._SubParsersAction) -> tuple[argparse._SubParsersAction, Any]:
    cmd = subparsers.add_parser("read_all_checks", help="Retrieve and print all checks from the data store")

    async def fun(ctx: ChecksRunnerContext, x: NamedArgs):
        # AFAIK, `enumerate`, nor `itertools` can handle an async iterator, so we enumerate manually
        c = 0
        async for check in await ctx.checks.read_all():
            c += 1
            print(f"{util.str_pad(c)}: {check}")

    cmd.set_defaults(fun=fun)

    return (subparsers, cmd)


_add_read_all_checks(SUBPARSERS)


# -----------------------------------------------------------------------------


def _add_delete_check(subparsers: argparse._SubParsersAction) -> tuple[argparse._SubParsersAction, Any]:
    cmd = subparsers.add_parser(
        "delete_check",
        help="Delete a check from the data store",
    )
    cmd.add_argument("url", **_url_kwargs(help="The check's URL to delete"))

    async def fun(ctx: ChecksRunnerContext, x: NamedArgs):
        print(await ctx.checks.delete(x.url))

    cmd.set_defaults(fun=fun)

    return (subparsers, cmd)


_add_delete_check(SUBPARSERS)


# -----------------------------------------------------------------------------


def _add_delete_all_checks(subparsers: argparse._SubParsersAction) -> tuple[argparse._SubParsersAction, Any]:
    cmd = subparsers.add_parser(
        "delete_all_checks",
        help="Delete all checks from the data store (use with caution)",
    )
    cmd.add_argument(
        "--confirm", help="For safety, you must activate this flag to delete all checks", action="store_true"
    )

    async def fun(ctx: ChecksRunnerContext, x: NamedArgs):
        ret = await ctx.checks.delete_all(x.confirm)
        print("done" if ret < 0 else ret)

    cmd.set_defaults(fun=fun)

    return (subparsers, cmd)


_add_delete_all_checks(SUBPARSERS)


# -----------------------------------------------------------------------------


def _add_check_website_only(subparsers: argparse._SubParsersAction) -> tuple[argparse._SubParsersAction, Any]:
    cmd = subparsers.add_parser(
        "check_website_only",
        help="Check given single website once (without writing to the data store)",
    )
    cmd.add_argument("url", **_url_kwargs())
    cmd.add_argument("--regex", **_regex_kwargs())

    async def fun(ctx: ChecksRunnerContext, x: NamedArgs):
        result = await ctx.check_only(WebsiteCheck.with_validation(x.url, x.regex))
        print(result)

    cmd.set_defaults(fun=fun)

    return (subparsers, cmd)


_add_check_website_only(SUBPARSERS)


# -----------------------------------------------------------------------------


def _add_check_website(subparsers: argparse._SubParsersAction) -> tuple[argparse._SubParsersAction, Any]:
    cmd = subparsers.add_parser(
        "check_website", help="Check given single website once and write the result in the data store"
    )
    cmd.add_argument("url", **_url_kwargs())
    cmd.add_argument("--regex", **_regex_kwargs())

    async def fun(ctx: ChecksRunnerContext, x: NamedArgs):
        result = await ctx.check_only(WebsiteCheck.with_validation(x.url, x.regex))
        await ctx.results.write(result)
        print(result)

    cmd.set_defaults(fun=fun)

    return (subparsers, cmd)


_add_check_website(SUBPARSERS)


# -----------------------------------------------------------------------------


def _check_all_once(subparsers: argparse._SubParsersAction) -> tuple[argparse._SubParsersAction, Any]:
    cmd = subparsers.add_parser(
        "check_all_once",
        help="Check all websites once and write the results in the data store (without scheduling; you might want to schedule this command with crontab)",
    )

    async def fun(ctx: ChecksRunnerContext, x: NamedArgs):
        c = 0
        async for check in ctx.check_all_once_n_write():
            c += 1
            print(f"{util.str_pad(c)}: {check}")

    cmd.set_defaults(fun=fun)

    return (subparsers, cmd)


_check_all_once(SUBPARSERS)


# -----------------------------------------------------------------------------


def _check_all_loop_fg(subparsers: argparse._SubParsersAction) -> tuple[argparse._SubParsersAction, Any]:
    cmd = subparsers.add_parser(
        "check_all_loop_fg",
        help=f"Check all websites in the foreground at the scheduled intervals (or {conf.DEFAULT_CHECK_INTERVAL_SECONDS}s for checks without an interval)",
    )

    async def fun(ctx: ChecksRunnerContext, x: NamedArgs):
        await ctx.run_checks_until_stopped_in_foreground()

    cmd.set_defaults(fun=fun)

    return (subparsers, cmd)


_check_all_loop_fg(SUBPARSERS)


# -----------------------------------------------------------------------------


def _add_read_last_results(subparsers: argparse._SubParsersAction) -> tuple[argparse._SubParsersAction, Any]:
    _DEFAULT_READ_N_RESULTS = 10

    cmd = subparsers.add_parser("read_last_results", help="Read the last results from the data store")
    cmd.add_argument(
        "-n",
        type=vutil.validated_parsed_is_positive_int,
        help=f"(Default: {_DEFAULT_READ_N_RESULTS}) The number of results to read",
        default=_DEFAULT_READ_N_RESULTS,
    )

    async def fun(ctx: ChecksRunnerContext, x: NamedArgs):
        print("(last results first)")
        c = 0
        async for result in ctx.results.read_last_n(x.n):
            c += 1
            print(f"{util.str_pad(c)}: {result}")

    cmd.set_defaults(fun=fun)

    return (subparsers, cmd)


_add_read_last_results(SUBPARSERS)


# -----------------------------------------------------------------------------


def parse_args(argv: list[str]) -> NamedArgs:
    args = PARSER.parse_args(argv)

    if args.command is None:
        print("Error: you must specify a command")
        sys.exit(2)

    return args


def parse_str_args(argv: str) -> NamedArgs:
    return parse_args(argv.split())


async def main(args: NamedArgs) -> None:
    # args must and are assumed to be validated

    async with ChecksRunnerContext.init_with_postgres(conf.get_postgres_conninfo()) as ctx:
        await args.fun(ctx, args)


# -----------------------------------------------------------------------------

if __name__ == "__main__":
    # ignore the first argument, which is the program name/path
    args = parse_args(sys.argv[1:])

    try:
        asyncio.run(main(args))
    except (KeyboardInterrupt, SystemExit):
        # ignore program-exit-like exceptions in the cli
        pass



import logging
import re2
import aiohttp

from fastchecks import conf
from fastchecks.types import CheckResult, WebsiteCheck
from fastchecks.util import (
    get_utcnow,
    get_utcnow_time_difference_seconds,
    is_likely_text_based_body,
    is_content_length_less_than,
)


async def check_website(
    session: aiohttp.ClientSession, check: WebsiteCheck, timeout: float | None = None
) -> CheckResult:
    """
    Access (GET) website's URL and return monitoring statistics (e.g. return status, response time, etc.).

    The information to check is encapsulated in the input WebsiteCheck instance.
    It's ASSUMED (but not asserted) that the input WebsiteCheck instance is valid.
    Make sure to have validated the inputs before calling this method (e.g. using WebsiteCheck.create_with_validation()).

    Optionally: if check.regex is defined, we check if the website's text body matches the regex.
    """
    timestamp_start = get_utcnow()

    _timeout = conf.DEFAULT_REQ_TIMEOUT_SECONDS if timeout is None else timeout

    # Note: if the input regex is None, theoretically we could do a HEAD request instead of a GET
    # However, often websites do not support HEAD, so we stick to GET
    response_ftr = session.get(check.url, timeout=_timeout)

    try:
        response = await response_ftr

        regex_match = (
            None
            if (
                # Note: when the response is not OK (<400), we do not check the regex
                not response.ok
                or check.regex is None
            )
            else await search_pattern_whole_text_body(check.regex, response)
        )

        # Get response time after (optionally) fetching the website's content (i.e., if the input regex is not None)
        response_time = get_utcnow_time_difference_seconds(timestamp_start)

        return CheckResult.response(
            check,
            timestamp_start,
            response_time,
            response_status=response.status,
            regex_match=regex_match,
        )

    except Exception as e:
        response_time = get_utcnow_time_difference_seconds(timestamp_start)

        match e:
            case TimeoutError():
                return CheckResult.failure(
                    check,
                    timestamp_start,
                    response_time,
                    timeout_error=True,
                )
            case aiohttp.ClientConnectorError():
                logging.debug(f"{e}")  # nothing major, it can happen

                return CheckResult.failure(
                    check,
                    timestamp_start,
                    response_time,
                    host_error=True,
                )
            case _:
                # unregistered exception, we log it
                logging.warn(f"UNKNOWN EXCEPTION: {e}", exc_info=True)

                return CheckResult.failure(
                    check,
                    timestamp_start,
                    response_time,
                    other_error=True,
                )

    finally:
        response_ftr.close()


async def search_pattern_whole_text_body(regex: str, response: aiohttp.ClientResponse) -> str | bool | None:
    """
    Search for a regex pattern in the response's content (assumed to be in most cases HTML).

    WARNING: the whole response's body is read in memory.

    To alleviate this:
    * we only read the response's body if it's likely to be text based (in particular, not binary) and
    * the response's Content-Length header is None (note: some websites do not report it) or is less than `__ARBITRARY_TOO_BIG_CONTENT_LENGTH`.
    -
    If, according to the rules above, the response's body is not read, we return None. Thus, the regex does not get tested.


    MAYBE (Alternatives):
    * If regex search can limited to a line, we could use use response.content.readline() instead of response.text().
    * The text body searched is raw HTML (in most cases), not the HTML's text. If we want to search the text of the HTML (or other text-based format) only, we would need a corresponding parser.
    """
    if is_likely_text_based_body(response) and is_content_length_less_than(
        response, length=conf.TOO_BIG_CONTENT_LENGTH_KB, allow_none_content_length=True
    ):
        content = await response.text()
        match_opt = re2.search(regex, content)

        if match_opt:
            return match_opt[0]
        else:
            return False
    else:
        logging.warning(
            f"The regex will not be checked because the response's body might be unsafe to read in memory (too big or not text-based), for url: {response.url}"
        )
        return None



__REQUIREMENT_FAILED_MSG_SUFFIX = "requirement failed"


def require(condition: bool, message: str | None = None) -> None:
    """
    Raise ValueError if condition is False, ala scala's `require` function.

    Safe alternative to `assert` to verify value input requirements.
    Note that the statement `assert` is not executed in production by default.
    See:
    * https://bandit.readthedocs.io/en/latest/plugins/b101_assert_used.html#b101-test-for-use-of-assert
    * https://snyk.io/blog/the-dangers-of-assert-in-python/
    """
    if not condition:
        msg = __REQUIREMENT_FAILED_MSG_SUFFIX if message is None else f"{__REQUIREMENT_FAILED_MSG_SUFFIX}: {message}"
        raise ValueError(msg)



import logging
from typing import AsyncIterator
from pydantic import PositiveInt
from fastchecks.types import CheckResult, WebsiteCheckScheduled, WebsiteCheck
from fastchecks.sockets import CheckResultSocket, WebsiteCheckSocket
from psycopg_pool import AsyncConnectionPool
from psycopg.rows import namedtuple_row
from psycopg import sql


class WebsiteCheckSocketPostgres(WebsiteCheckSocket):
    def __init__(self, conninfo: str) -> None:
        self.__pool = AsyncConnectionPool(conninfo)

    def is_closed(self) -> bool:
        return self.__pool.closed

    async def upsert(self, check: WebsiteCheckScheduled) -> None:
        async with self.__pool.connection() as aconn:
            await aconn.execute(
                """
            INSERT INTO WebsiteCheck
            (url, regex, interval_seconds)
            VALUES (%s, %s, %s)
            ON CONFLICT (url) DO UPDATE
                SET regex = EXCLUDED.regex,
                    interval_seconds = EXCLUDED.interval_seconds;
            """,
                (check.url, check.regex, check.interval_seconds),
            )

    async def read_n(self, n: PositiveInt) -> AsyncIterator[WebsiteCheckScheduled]:
        async with self.__pool.connection() as aconn:
            # We format the safe query in 2 steps (also below) to avoid false positives from bandit (B608:hardcoded_sql_expressions)
            # We follow psycopg recommendations for how to escape composed SQL queries: https://www.psycopg.org/psycopg3/docs/advanced/typing.html#checking-literal-strings-in-queries
            query_raw = """
                SELECT * FROM WebsiteCheck
                LIMIT {};"""
            query_safe = query_raw.format(n)

            acur = await aconn.execute(sql.SQL(query_safe))

            acur.row_factory = namedtuple_row
            async for row in acur:
                # without validation, because we trust the database -- its value were validated before
                yield WebsiteCheckScheduled.with_check(
                    WebsiteCheck.without_validation(row.url, row.regex), row.interval_seconds
                )

    async def delete(self, url: str) -> int:
        async with self.__pool.connection() as aconn:
            cur = await aconn.execute(
                """
            DELETE FROM WebsiteCheck
            WHERE url = %s;""",
                (url,),
            )
            return cur.rowcount

    async def delete_all(self, confirm: bool) -> int:
        if confirm:
            async with self.__pool.connection() as aconn:
                cur = await aconn.execute(
                    """
                TRUNCATE WebsiteCheck;"""
                )
                return cur.rowcount
        else:
            logging.warning("Not deleting all checks because confirm is False.")
            return 0

    async def close(self) -> None:
        return await self.__pool.close()


class CheckResultSocketPostgres(CheckResultSocket):
    def __init__(self, conninfo: str) -> None:
        self.__pool = AsyncConnectionPool(conninfo)

    def is_closed(self) -> bool:
        return self.__pool.closed

    async def write(self, result: CheckResult) -> None:
        async with self.__pool.connection() as aconn:
            await aconn.execute(
                """
            INSERT INTO CheckResult
            (url, regex, timestamp_start, response_time, timeout_error, host_error, other_error, response_status, regex_match)
            VALUES (%s, %s, %s, %s, %s, %s, %s, %s, %s);""",
                (
                    result.check.url,
                    result.check.regex,
                    #
                    result.timestamp_start,
                    result.response_time,
                    #
                    result.timeout_error,
                    result.host_error,
                    result.other_error,
                    #
                    result.response_status,
                    result.regex_match_to_bool_or_none(),
                ),
            )

    async def read_last_n(self, n: PositiveInt) -> AsyncIterator[CheckResult]:
        async with self.__pool.connection() as aconn:
            query_raw = """
                SELECT * FROM CheckResult
                ORDER BY timestamp_start DESC
                LIMIT {};"""
            query_safe = query_raw.format(n)

            acur = await aconn.execute(sql.SQL(query_safe))

            acur.row_factory = namedtuple_row
            async for row in acur:
                yield CheckResult(
                    check=WebsiteCheck.without_validation(row.url, row.regex),
                    #
                    timestamp_start=row.timestamp_start,
                    response_time=row.response_time,
                    #
                    timeout_error=row.timeout_error,
                    host_error=row.host_error,
                    other_error=row.other_error,
                    #
                    response_status=row.response_status,
                    regex_match=row.regex_match,
                )

    async def close(self) -> None:
        return await self.__pool.close()
